;; Generated package description from map.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "map" "3.3.1.0.20260414.164123" "Map manipulation functions" '((emacs "26")) :commit "0384600a21fb363e623afa658d1e66a54f79d2fb" :url "https://elpa.gnu.org/packages/map.html" :authors '(("Nicolas Petton" . "nicolas@petton.fr")) :maintainer '(nil . "emacs-devel@gnu.org") :keywords '("extensions" "lisp"))
